/*
 * control_actions.hpp
 *
 * Copyright (c) Stephen Thompson, 2007.
 * Licenced for non-commercial use only. See LICENCE.txt for details.
 *
 * Actions that make a knight do something. These are typically used
 * for Controls.
 *
 */

#ifndef CONTROL_ACTIONS_HPP
#define CONTROL_ACTIONS_HPP

#include "action.hpp"

//
// Attack, Move, Withdraw and Suicide are special in that they are handled implicitly
// rather than being passed to DungeonView::setAvailableControls in the usual way.
// They also can't be created in the config system at present. (To allow this, would have
// to add possible() functions I guess.)
//

void SetupStandardControls();

class A_Attack : public Action {
public:
	explicit A_Attack(MapDirection d) : dir(d) { }
	virtual void execute(const ActionData &) const;
private:
	MapDirection dir;
};

class A_Move : public Action {
public:
	explicit A_Move(MapDirection d) : dir(d) { }
	virtual void execute(const ActionData &) const;
private:
	MapDirection dir;
};

class A_Withdraw : public Action {
public:
	virtual void execute(const ActionData &) const;
};

class A_CancelMove : public Action {
public:
	virtual bool possible(const ActionData &) const;
	virtual void execute(const ActionData &) const;
};

class A_Suicide : public Action {
public:
	virtual void execute(const ActionData &) const;
};


//
// Other control-related actions.
//

class A_Drop : public Action {
public:
	explicit A_Drop(const ItemType &it_) : it(it_) { }
	virtual bool possible(const ActionData &) const;
	virtual void execute(const ActionData &) const;
private:
	const ItemType &it;
	ACTION_MAKER("Drop");
};

class A_DropHeld : public Action {
public:
	virtual bool possible(const ActionData &) const;
	virtual void execute(const ActionData &) const;
private:
	ACTION_MAKER("DropHeld");
};

class A_PickLock : public Action {
public:
	A_PickLock(int p, int wt) : prob(p), waiting_time(wt) { }
	virtual bool possible(const ActionData &) const;
	virtual void execute(const ActionData &) const;
private:
	int prob, waiting_time; // prob in %, waiting_time in ticks
	ACTION_MAKER("PickLock");
};

class A_PickUp : public Action {
public:
	virtual bool possible(const ActionData &) const;
	virtual void execute(const ActionData &) const;
private:
	ACTION_MAKER("PickUp");
};

class A_SetBearTrap : public Action {
public:
	explicit A_SetBearTrap(const ItemType &trap_) : trap_itype(trap_) { }
	virtual bool possible(const ActionData &) const;
	virtual void execute(const ActionData &) const;
private:
	const ItemType &trap_itype;
	ACTION_MAKER("SetBearTrap");
};

class A_SetBladeTrap : public Action {
public:
	explicit A_SetBladeTrap(const ItemType &missile) : missile_type(missile) { }
	virtual bool possible(const ActionData &) const;
	virtual void execute(const ActionData &) const;
private:
	const ItemType &missile_type;
	ACTION_MAKER("SetBladeTrap");
};

class A_SetPoisonTrap : public Action {
public:
	virtual bool possible(const ActionData &) const;
	virtual void execute(const ActionData &) const;
private:
	ACTION_MAKER("SetPoisonTrap");
};

class A_Swing : public Action {
public:
	virtual bool possible(const ActionData &) const;
	virtual void execute(const ActionData &) const;
private:
	ACTION_MAKER("Swing");
};

// A_Throw is usually used for throwing daggers. It looks at the
// ActionData Item parameter to determine which item to throw.
class A_Throw : public Action {
public:
	virtual bool possible(const ActionData &) const;
	virtual void execute(const ActionData &) const;
private:
	ACTION_MAKER("Throw");
};

#endif
