/*
 * creature.cpp
 *
 * Copyright (c) Stephen Thompson, 2007.
 * Licenced for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "anim.hpp"
#include "cfg.hpp"
#include "creature.hpp"
#include "dungeon_map.hpp"
#include "item_type.hpp"
#include "knight.hpp"
#include "mediator.hpp"
#include "missile.hpp"
#include "rng.hpp"
#include "task.hpp"
#include "task_manager.hpp"

//
// ImpactTask -- this processes melee weapon impacts.
//

class ImpactTask : public Task {
public:
	ImpactTask(weak_ptr<Creature> c_) : c(c_) { }
	virtual void execute(TaskManager &);
private:
	enum ImpactResult { NO_TARGET, HIT, VETO };
	ImpactResult tryCreatureImpact(Creature &me);
private:
	weak_ptr<Creature> c;
};

ImpactTask::ImpactResult ImpactTask::tryCreatureImpact(Creature &me)
{
	// Find target
	if (!me.getMap()) return NO_TARGET;
	shared_ptr<Creature> target_creature = me.getMap()->getTargetCreature(me, true);
	if (!target_creature) return NO_TARGET;
	
	// The "impactVeto" process is designed to ensure that
	// there is a fair result when two creatures strike each
	// other simultaneously (or near simultaneously).
	const int gvt = Mediator::instance().getGVT();
	if (target_creature->getPos() != me.getPos() && target_creature->getFacing()
	== Opposite(me.getFacing()) && target_creature->impactVeto(gvt, me)) {
		// enemy is going to hit me first -- abort my attack and wait.
		me.setAnimFrame(0, 0);
		me.stunUntil(gvt + CfgInt("attack_threshold") + 1);
		return VETO;
	}

	// Parry Check:
	if (target_creature->getFacing() == Opposite(me.getFacing())
		&& !target_creature->isApproaching()
		&& !target_creature->isMoving()
		&& !target_creature->stunned()) {
		const ItemType *parry_weapon = target_creature->getItemInHand();
		if (parry_weapon) {
			const float parry_chance = parry_weapon->getParryChance() / 100.0f;
			if (g_rng.getBool(parry_chance)) {
				// Successful parry
				target_creature->parry();
				return HIT;
			}
		}
	}
				
	// If we get to here then we may proceed to hit the target.
	if (me.getItemInHand())
		me.getItemInHand()->doCreatureImpact(gvt,
				static_pointer_cast<Creature>(me.shared_from_this()),
				target_creature, me.hasStrength());
	return HIT;
}	

void ImpactTask::execute(TaskManager &tm)
{
	shared_ptr<Creature> me(c.lock());
	if (!me) return;
	if (!me->item_in_hand) return;  // attacker doesn't seem to have a weapon
	if (me->isMoving()) return;     // we should not be able to swing weapon while moving...

	const int gvt = tm.getGVT();
	if (me->impact_time != gvt) return;  // this is the wrong impact task, or impact was
										 // cancelled.
	me->impact_time = 0;

	DungeonMap * dmap = me->getMap();
	if (!dmap) return;

	// Run the downswing hook.
	const char *hook = me->getWeaponDownswingHook();
	if (hook) {
		Mediator::instance().runHook(hook, me);
	}
	
	// Check impacts against creatures -- first in the tile ahead, then in my current tile
	// (the latter makes it possible to hit vampire bats while they are flying over you,
	// which makes them a little easier to kill).
	ImpactResult result = tryCreatureImpact(*me);

	if (result == NO_TARGET) {
		// Couldn't find any creatures to hit; hit the tile itself.
		me->item_in_hand->doTileImpact(me, *dmap, DisplaceCoord(me->getPos(),me->getFacing()),
									   me->hasStrength());
	}

	// Clean up attacker's anim state and stun time (assuming impact was not vetoed).
	if (result != VETO) {
		const int qf = me->hasQuickness() ? CfgInt("quickness_factor") : 100;
		const int ds_time = me->item_in_hand->getMeleeDownswingTime();
		const int wt_time = ds_time + CfgInt("melee_delay_time");
		me->setAnimFrame(AF_IMPACT, gvt + ds_time * 100 / qf);
		me->stunUntil(gvt + wt_time * 100 / qf);
	}
}

//
// Creature::impactVeto
//

bool Creature::impactVeto(int gvt, const Creature &attacker)
{
	if (impact_time >= gvt && impact_time <= gvt + CfgInt("attack_threshold")) {
		// I am about to hit the attacker, just as he hits me, ie the
		// impact times are simultaneous (to within attack_threshold).
		// In this case we decide 50/50 who wins.

		// SPECIAL RULE: if the attacker is a monster (non-knight creature) and the defender
		// is a knight, then the impact is always vetoed. (This makes the zombies slightly
		// easier to kill, and it matches the original Knights behaviour better.)
		// Conversely, if the attacker is a knight and the defender is a monster, then the
		// impact can never be vetoed.

		const bool attacker_is_knight = dynamic_cast<const Knight *>(&attacker);
		const bool defender_is_knight = dynamic_cast<const Knight *>(this);
		const bool must_veto = !attacker_is_knight && defender_is_knight;
		const bool must_not_veto = attacker_is_knight && !defender_is_knight;
		if (must_veto || (!must_not_veto && g_rng.getBool(0.5f))) {
			// attacker loses (veto his impact)
			return true;
		} else {
			// I lose -- cancel my own impact.
			setAnimFrame(0, 0);
			stunUntil(gvt + CfgInt("attack_threshold"));
			return false;
		}
	} else {
		// No simultaneous impact ("normal" case)
		return false; // no veto
	}
}


//
// ThrowingTask -- throws a missile weapon (after backswing has completed).
//

class ThrowingTask : public Task {
public:
	ThrowingTask(weak_ptr<Creature> c_, const ItemType &i) : c(c_), itype(i) { }
	virtual void execute(TaskManager &);
private:
	weak_ptr<Creature> c;
	const ItemType &itype;
};

void ThrowingTask::execute(TaskManager &tm)
{
	shared_ptr<Creature> me(c.lock());
	if (!me) return;
	if (me->isMoving()) return;

	const int gvt = tm.getGVT();
	if (me->impact_time != gvt) return;
	me->impact_time = 0;

	DungeonMap *dmap = me->getMap();
	if (!dmap) return;

	// create the missile
	bool success = CreateMissile(*dmap, me->getPos(), me->getFacing(),
								 itype, true, me->hasStrength());

	// reset anim / stun time
	const int qf = me->hasQuickness() ? CfgInt("quickness_factor") : 100;
	const int ds_time = itype.getMissileDownswingTime();
	const int wt_time = ds_time + CfgInt("melee_delay_time");
	me->setAnimFrame(AF_THROW_DOWN, gvt + ds_time * 100 / qf);
	me->stunUntil(gvt + wt_time * 100 / qf);

	// get rid of the item (if it was successfully thrown).
	if (success) me->throwAwayItem(&itype);
	else me->throwAwayItem(0);
}


//
// Creature
//

Creature::Creature(int h, MapHeight ht, const ItemType * i, const Anim * a, int s)
	: height(ht), health(h), max_health(h), speed(s), stun_until(0), item_in_hand(i)
{
	// Hmm, perhaps "speed" should be in Knight rather than Creature.
	setAnim(a);
	setSpeed(speed);
	if (i) setOverlay(i->getOverlay());
}


bool Creature::stunned() const
{
	if (!getMap()) return false;
	return (stun_until > Mediator::instance().getGVT() || isParalyzed());
}

bool Creature::canAct() const
{
	return getMap() && !stunned() && !isMoving();
}

void Creature::stunUntil(int new_stun_time)
{
	if (new_stun_time > stun_until) stun_until = new_stun_time;
	if (impact_time && stunned()) impact_time = 0; // cancel impact-in-progress if stunned.
}

void Creature::stunnedUntil(bool &known, int &s_until) const
{
	if (!getMap()) {
		// Creature isn't in map -- we cannot access GVT.
		known = false;
	} else if (isParalyzed()) {
		// If we are paralyzed, then no lower bound can be given, since there could
		// be a DISPEL MAGIC at any time.
		known = false;
	} else if (stun_until > Mediator::instance().getGVT()) {
		// We know that definitely we will be stunned until stun_until
		known = true;
		s_until = stun_until;
	} else {
		// Hmm, it doesn't look like we are stunned after all.
		// Set known to false in this case
		known = false;
	}
}

void Creature::cannotActUntil(bool &known, int &ca_until) const
{
	// Cannot act = Stunned OR Moving
	stunnedUntil(known, ca_until);
	if (isMoving()) {
		ASSERT(getArrivalTime() != 0);
		ASSERT(getArrivalTime() > Mediator::instance().getGVT());
		const int at = getArrivalTime();
		if (!known || at > ca_until) {
			known = true;
			ca_until = at;
		}
	}
}


void Creature::doSetItemInHand(const ItemType * i)
{
	item_in_hand = i;
	if (i) setOverlay(i->getOverlay());
	else setOverlay(0);
}

//
// melee
//

bool Creature::canSwing() const
{
	if (!item_in_hand) return false;
	if (!item_in_hand->canSwing()) return false;
	if (!canAct()) return false;
	if (isMoving() || isApproaching()) return false;  // can only swing while standing in centre of a sq.
	return true;
}

void Creature::swing()
{
	if (!getMap()) return;
	if (!canSwing()) return;
	TaskManager &tm(Mediator::instance().getTaskManager());
	
	const int gvt = tm.getGVT();
	const int qf = hasQuickness() ? CfgInt("quickness_factor") : 100;
	this->impact_time = gvt + item_in_hand->getMeleeBackswingTime() * 100 / qf;

	// set anim and stun
	if (impact_time > stun_until) stun_until = impact_time;
	setAnimFrame(AF_BACKSWING, impact_time + 1);

	// set the impact task.
	shared_ptr<ImpactTask> imp_tsk(new ImpactTask(static_pointer_cast<Creature>(
		shared_from_this())));
	tm.addTask(imp_tsk, TP_NORMAL, impact_time);
}

//
// throwing
//

bool Creature::canThrow(bool strict) const
{
	if (!getItemInHand()) return false;
	if (!getItemInHand()->canThrow()) return false;
	return canDoThrow(strict);
}

void Creature::throww()
{
	if (!canThrow(true)) return;
	if (!getItemInHand()) return;
	doThrow(*getItemInHand());
}

bool Creature::canDoThrow(bool strict) const
{
	if (strict && !canAct()) return false;
	if (isMoving() || isApproaching()) return false;
	const DungeonMap *dmap = getMap();
	if (!dmap) return false;
	if (strict && dmap->getAccess( DisplaceCoord(getPos(), getFacing()),
		MapHeight(H_MISSILES + getFacing()) ) != A_CLEAR) return false;
	return true;
}

void Creature::doThrow(const ItemType &itype)
{
	if (!getMap()) return;
	if (!canAct()) return;
	TaskManager &tm(Mediator::instance().getTaskManager());
	
	// Work out timings.
	const int gvt = tm.getGVT();
	const int qf = hasQuickness() ? CfgInt("quickness_factor") : 100;
	this->impact_time = gvt + itype.getMissileBackswingTime() * 100 / qf;

	// Set anim and stun (for backswing).
	if (impact_time > stun_until) stun_until = impact_time;
	setAnimFrame(AF_THROW_BACK, impact_time + 1);

	// The actual throw is done by a ThrowingTask (once the backswing completes).
	shared_ptr<ThrowingTask> thr_tsk(new
		ThrowingTask(static_pointer_cast<Creature>(shared_from_this()), itype));
	tm.addTask(thr_tsk, TP_NORMAL, impact_time);
}

void Creature::throwAwayItem(const ItemType *itype)
{
	// This routine gets rid of an item that was just thrown.
	// Default implementation assumes that the item is currently held.
	// (Knight overrides this routine to look in the backpack as well.)
	if (itype == getItemInHand()) setItemInHand(0);
}

//
// parrying
//

void Creature::parry()
{
	if (!getMap()) return;
	const int finish_time = Mediator::instance().getGVT() + CfgInt("parry_delay");
	setAnimFrame(AF_PARRY, finish_time);
	stunUntil(finish_time);

	// run hook
	Mediator::instance().runHook("HOOK_WEAPON_PARRY", static_pointer_cast<Creature>(shared_from_this()));
}


//
// damage and hitpoints
//

void Creature::damage(int amount, int su /* = -1 */)
{
	if (amount > 0) {
		health -= amount;

		if (getMap()) {
			bool need_blood = true;
			const int blood_level = bloodLevel();
			if (blood_level==0) need_blood = false;
			if (blood_level==1 && health > 0) need_blood = false;
			if (need_blood) Mediator::instance().placeBlood(*getMap(), getNearestPos());
			if (health <= 0) {
				// Death!
				// call onDeath() (this causes knights to drop items, also does corpses);
				// remove myself from the map (this may also trigger on_withdraw events).
				onDeath(NORMAL_MODE);
				rmFromMap();
				return;  // no point continuing if creature is dead :)
			}
		}
	}
	if (su != -1) stunUntil(su);
}

void Creature::poison()
{
	// death by poison.
	if (getMap()) {
		onDeath(POISON_MODE);
		rmFromMap();
	}
}

void Creature::addToHealth(int amount)
{
	if (amount > 0) {
		health += amount;
		if (health > max_health) health = max_health;
	}
}

