/*
 * remote_dungeon_view.cpp
 *
 * Copyright (c) Stephen Thompson, 2007.
 * Licenced for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "anim.hpp"
#include "binary_io/binary_io.hpp"
#include "command_codes.hpp"
#include "control.hpp"
#include "misc/my_exceptions.hpp"
#include "platform/graphic.hpp"
#include "remote_dungeon_view.hpp"
#include "server_config.hpp"
#include "task_manager.hpp"

#include <iostream>  // for inheritance relationships (eg iostream is-an ostream)
using namespace std;

//
// "send" functions
//

void RemoteDungeonView::sendNibbles(int x, int y)
{
	if (x < 0 || x > 15 || y < 0 || y > 15) {
		throw UnexpectedError("RDV: nibble out of range");
	}
	BinWriteUbyte(str, (x<<4) + y);
}

void RemoteDungeonView::sendUbyte(int x)
{
	if (x < 0 || x > 255) {
		throw UnexpectedError("RDV: ubyte out of range");
	}
	BinWriteUbyte(str, x);
}

void RemoteDungeonView::sendUshort(int x)
{
	if (x < 0 || x > 65535) {
		throw UnexpectedError("RDV: ushort out of range");
	}
	BinWriteUword(str, x);
}

void RemoteDungeonView::sendLong(int x)
{
	BinWriteLong(str, x);
}

void RemoteDungeonView::sendAnimPtr(const Anim *x)
{
	sendUbyte(x? x->getID() : 0);
}

void RemoteDungeonView::sendColourChange(const ColourChange &cc)
{
	cc.serialize(str);
}

void RemoteDungeonView::sendControlPtr(const Control *c)
{
	sendUbyte(c? c->getID() : 0);
}

void RemoteDungeonView::sendControlPlusBool(const Control *c, bool b)
{
	sendUbyte((c? c->getID() : 0) + 128*int(b));
}

void RemoteDungeonView::sendGraphicPtr(const Graphic *g)
{
	sendUshort(g? GetID(*g) : 0);
}

void RemoteDungeonView::sendOverlayPtr(const Overlay *x)
{
	sendUbyte(x? x->getID() : 0);
}

void RemoteDungeonView::sendString(const string &s)
{
	// TODO: We might improve this by using a hash table of previously sent strings.
	BinWriteString(str, s);
}

void RemoteDungeonView::sendMapCoord(const MapCoord &mc)
{
	if (mc.isNull()) {
		sendUbyte(255);
	} else {
		if (mc.getX()>=255 || mc.getY()>=255 || mc.getX()<0 || mc.getY()<0) {
			throw UnexpectedError("map coord out of range");
		}
		sendUbyte(mc.getX());
		sendUbyte(mc.getY());
	}
}


//
// sendCommandNo -- also updates the time if necessary
//

void RemoteDungeonView::sendCommandNo(int x)
{
	sendUbyte(x);
}


//
// functions to send commands to the client
//

void RemoteDungeonView::setCurrentRoom(int r, int w, int h)
{
	if (game_over) return;
	
	sendCommandNo(C_SET_CURRENT_ROOM);
	sendRoomNo(r);
	sendRoomCoord(w-1, h-1);
	
	// we also store the current room for use by getCurrentRoom
	current_room = r;
	cur_room_width = w;
	cur_room_height = h;
}

int RemoteDungeonView::getCurrentRoom() const
{
	return current_room;
}

void RemoteDungeonView::mapCurrentRoom(const MapCoord &bl_pos)
{
	if (game_over) return;
	sendCommandNo(C_MAP_CURRENT_ROOM);
	sendMapCoord(bl_pos);

	// update mapped_squares
	if (current_room >= 0) {
		// Dynamically resize the array if necessary
		const int sx = max(bl_pos.getX() + cur_room_width, int(mapped_squares.shape()[0]));
		const int sy = max(bl_pos.getY() + cur_room_height, int(mapped_squares.shape()[1]));
		mapped_squares.resize(boost::extents[sx][sy]);
		for (int y=bl_pos.getY(); y < bl_pos.getY() + cur_room_height; ++y) {
			for (int x=bl_pos.getX(); x < bl_pos.getX() + cur_room_width; ++x) {
				mapped_squares[x][y] = true;
			}
		}
	}
}

void RemoteDungeonView::mapWall(const MapCoord &mc)
{
	if (game_over) return;
	sendCommandNo(C_MAP_WALL);
	sendMapCoord(mc);
	// (For simplicity, we do not update mapped_squares in this case. We ought to really,
	// but it doesn't matter because isSquareMapped is never used on walls, at time of
	// writing.)
}

void RemoteDungeonView::mapKnightLocation(int n, const MapCoord &mc)
{
	if (game_over) return;	
	sendCommandNo(C_MAP_KNIGHT_LOCATION);
	sendUbyte(n);
	sendMapCoord(mc);
}

void RemoteDungeonView::mapItemLocation(const MapCoord &mc, bool on)
{
	if (game_over) return;
	sendCommandNo(C_MAP_ITEM_LOCATION);
	sendMapCoord(mc);
	sendUbyte(on);
}

void RemoteDungeonView::wipeMap()
{
	if (game_over) return;
	sendCommandNo(C_WIPE_MAP);
	mapped_squares.resize(boost::extents[0][0]); // wipe mapped_squares too
}

bool RemoteDungeonView::isSquareMapped(const MapCoord &mc) const
{
	if (mc.getX() >= 0 && mc.getX() < mapped_squares.shape()[0]
	&& mc.getY() >= 0 && mc.getY() < mapped_squares.shape()[1]) {
		return mapped_squares[mc.getX()][mc.getY()];
	} else {
		return false;
	}
}

void RemoteDungeonView::addEntity(unsigned short int id, int x, int y, MapHeight ht, MapDirection facing,
								  const Anim *anim, const Overlay *ovr, int af, int atz_diff,
								  bool ainvuln, int cur_ofs, MotionType mt, int mtr)
{
	if (game_over) return;
	sendCommandNo(C_ADD_ENTITY);
	sendUshort(id);
	sendRoomCoord(x, y);
	sendNibbles(int(ht), facing);
	sendAnimPtr(anim);
	sendOverlayPtr(ovr);
	sendNibbles(af, (int(mt)<<1) + int(ainvuln));
	if (af != 0) BinWriteWord(str, atz_diff);
	sendUshort(cur_ofs);
	if (mt != MT_NOT_MOVING) sendUshort(mtr);
}

void RemoteDungeonView::rmEntity(unsigned short int id)
{
	if (game_over) return;
	sendCommandNo(C_RM_ENTITY);
	sendUshort(id);
}

void RemoteDungeonView::repositionEntity(unsigned short int id, int new_x, int new_y)
{
	if (game_over) return;
	sendCommandNo(C_REPOSITION_ENTITY);
	sendUshort(id);
	sendRoomCoord(new_x, new_y);
}

void RemoteDungeonView::moveEntity(unsigned short int id, MotionType motion_type,
								   int motion_duration)
{
	if (game_over) return;
	sendCommandNo(C_MOVE_ENTITY);
	sendUshort(id);
	sendNibbles(0, motion_type);
	sendUshort(motion_duration);
}

void RemoteDungeonView::flipEntityMotion(unsigned short int id, int motion_duration)
{
	if (game_over) return;
	sendCommandNo(C_FLIP_ENTITY_MOTION);
	sendUshort(id);
	sendUshort(motion_duration);
}

void RemoteDungeonView::setAnimData(unsigned short int id, const Anim *anim,
									const Overlay *ovr, int af, int atz_diff, bool ainvuln)
{
	if (game_over) return;
	sendCommandNo(C_SET_ANIM_DATA);
	sendUshort(id);
	sendAnimPtr(anim);
	sendOverlayPtr(ovr);
	sendNibbles(af, int(ainvuln));
	BinWriteWord(str, atz_diff);
}

void RemoteDungeonView::setFacing(unsigned short int id, MapDirection new_facing)
{
	if (game_over) return;
	sendCommandNo(C_SET_FACING);
	sendUshort(id);
	sendNibbles(0, new_facing);
}

void RemoteDungeonView::clearTiles(int x, int y)
{
	if (game_over) return;
	sendCommandNo(C_CLEAR_TILES);
	sendRoomCoord(x, y);
}

void RemoteDungeonView::setTile(int x, int y, int depth, const Graphic *gfx,
								shared_ptr<const ColourChange> cc, MiniMapColour c)
{
	if (game_over) return;
	sendCommandNo(C_SET_TILE);
	sendRoomCoord(x, y);
	sendTileInfo(depth, c, cc);
	sendGraphicPtr(gfx);
	if (cc) sendColourChange(*cc);
}

void RemoteDungeonView::setItem(int x, int y, const Graphic *g)
{
	if (game_over) return;
	sendCommandNo(C_SET_ITEM);
	sendRoomCoord(x, y);
	sendGraphicPtr(g);
}

void RemoteDungeonView::placeIcon(int x, int y, const Graphic *g, int dur)
{
	if (game_over) return;
	sendCommandNo(C_PLACE_ICON);
	sendRoomCoord(x, y);
	sendGraphicPtr(g);
	sendUshort(dur);
}

void RemoteDungeonView::setBackpack(int slot, const Graphic *g, int no_carried, int no_max)
{
	if (game_over) return;
	sendCommandNo(C_SET_BACKPACK);
	sendUbyte(slot);
	sendGraphicPtr(g);
	sendUbyte(no_carried);
	sendUbyte(no_max);
}

void RemoteDungeonView::addSkull()
{
	if (game_over) return;
	sendCommandNo(C_ADD_SKULL);
}

void RemoteDungeonView::setHealth(int h)
{
	if (game_over) return;
	sendCommandNo(C_SET_HEALTH);
	sendUbyte(max(0,h));
}

void RemoteDungeonView::setPotionMagic(PotionMagic pm, bool pi)
{
	if (game_over) return;
	sendCommandNo(C_SET_POTION_MAGIC);
	sendUbyte(128*int(pi) + int(pm));
}

void RemoteDungeonView::flashMessage(const string &msg)
{
	if (game_over) return;
	sendCommandNo(C_FLASH_MESSAGE);
	sendString(msg);
}

void RemoteDungeonView::cancelContinuousMessages()
{
	if (game_over) return;
	sendCommandNo(C_CANCEL_CONTINUOUS_MESSAGES);
}

void RemoteDungeonView::addContinuousMessage(const string &msg)
{
	if (game_over) return;
	sendCommandNo(C_ADD_CONTINUOUS_MESSAGE);
	sendString(msg);
}

void RemoteDungeonView::flashScreen(int delay)
{
	if (game_over) return;
	sendCommandNo(C_FLASH_SCREEN);
	sendUshort(delay);
}

void RemoteDungeonView::playSound(const string &name, int frequency)
{
	if (game_over) return;
	sendCommandNo(C_PLAY_SOUND);
	sendString(name); // TODO we should probably be caching recently sent strings ...
	sendUshort(frequency);
}

void RemoteDungeonView::winGame()
{
	if (game_over) return;	
	sendCommandNo(C_WIN_GAME);
	game_over = true;
	str.flush();
}

void RemoteDungeonView::loseGame()
{
	if (game_over) return;
	sendCommandNo(C_LOSE_GAME);
	game_over = true;
	str.flush();
}

void RemoteDungeonView::quitGame()
{
	if (game_over) return;
	sendCommandNo(C_QUIT_GAME);
	game_over = true;
	str.flush();
}

void RemoteDungeonView::setAvailableControls(const vector<pair<const Control*, bool> > &ctrls)
{
	if (game_over) return;	
	sendCommandNo(C_SET_AVAILABLE_CONTROLS);
	sendUbyte(ctrls.size());
	for (int i=0; i<ctrls.size(); ++i) {
		sendControlPlusBool(ctrls[i].first, ctrls[i].second);
	}
}

void RemoteDungeonView::setMenuHighlight(const Control *ctrl)
{
	if (game_over) return;
	sendCommandNo(C_SET_MENU_HIGHLIGHT);
	sendControlPtr(ctrl);
}

const Control * RemoteDungeonView::readController()
{
	if (game_over) return 0;
	const Control *c = control;
	while (str.rdbuf()->in_avail() > 0) {
		int x = BinReadUbyte(str);
		if (x == 255) {
			// 255 is the magic number sent when the game ends
			// (see also GameClient::sendGameOverMsg and InGameScreen::update).
			// We should not be receiving it here!
			throw UnexpectedError("network error: unexpected 'game over' message");
		}
		
		c = cfg.getControlFromID(x);
		if (c) {
			if (c->isContinuous()) control = c;
		} else {
			control = 0;
		}
	}
	return c;
}

void RemoteDungeonView::update()
{
	str.flush();
}
